/*
# Copyright (c) 2003-2004 by the University of Southern California.
# All rights reserved.
#
# Permission to use, copy, modify, and distribute this software and its
# documentation in source and binary forms for non-commercial purposes and
# without fee is hereby granted, provided that the above copyright notice
# appear in all copies and that both the copyright notice and this permission
# notice appear in supporting documentation, and that any documentation,
# advertising materials, and other materials related to such distribution and
# use acknowledge that the software was developed by the University of
# Southern California, Information Sciences Institute.  The name of the
# University may not be used to endorse or promote products derived from this
# software without specific prior written permission.
#
# THE UNIVERSITY OF SOUTHERN CALIFORNIA MAKES NO REPRESENTATIONS ABOUT THE
# SUITABILITY OF THIS SOFTWARE FOR ANY PURPOSE.  THIS SOFTWARE IS PROVIDED "AS
# IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT
# LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE.
#
# Other copyrights might apply to parts of this software and are so noted when
# applicable.
# ----------------------------------------------------------------------------
#
# Primary author: Purushotham Kamath
#                 Sept  2004
#This material is based upon work supported by the Defense Advanced Research 
#Projects Agency under contract no. N66001-02-1-8939 issued by the Space and 
#Naval Warfare Systems Center (SPAWAR).  Any opinions, findings, and conclusions 
#or recommendations expressed in this material are those of the author(s) and 
#do not necessarily reflect the views of the Defense Advanced Research Projects
#Agency, SPAWAR, or the U.S. Government.
*/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include <defs.h>
#include <packet.h>
#include <eventhandler.h>
#include <squeue.h>
#include <queue.h>
#include <scheduler.h>
#include <rng.h>
#include <trgen.h>
#include <debug.h>
#include <intransit.h>
#include <codeset.h>

/*
The InTransit object stores all packets that are currently being transmitted in the medium 
(i.e in the process of transmission , in transit or in the process of reception).
 Each packet object stored has the time that transmission began, the length of the packet 
and the point in space that the transmission began from. Given a specific type of topology 
object, and transmission coding parameters the position of any packet at any time may be determined.
Given the Intransit object, the state of the medium at any point in time or space may be determined by summing 
up the state due to each packet present at that point.

InTransit object is a singleton class
*/
extern TrGen * trgen;
extern Codeset * codeset;
extern int propSpeed;

InTransit::InTransit(int a, int n)
{
	q = new Queue();
	current = q->head;

}

InTransit::~InTransit()
{
	delete q;
}


void InTransit::add(Packet *p)
{

	//Store a copy of the packet in Intransit
	Packet *pt = new Packet(*p);
	q->enqueue(pt);	
	current = q->head;
	//dprintf("intransit.cc: Intransit::Add(): Dep time: %.12f\n", pt->departuretime);

}

void InTransit::remove(Packet *p)
{
	q->dequeue(p);
	current = q->head;
	delete p;
}

void InTransit::dump()
{
	Packet *p = q->head;
	while(p != NULL)
	{
		
		dprintf("Intransit: traverse: Packet: %d\n", p->id);
		p = p->next;
	}
}


Packet * InTransit::traverse()
{

	Packet * p;

	if(current == NULL)
	{

		//dprintf("intransit.cc: Intransit::traverse(): Found end\n");
		p = current;
		current =  q->head;
	}
	else
	{
		p = current;
		current = current->next;


//TBD: Removal algorithm	
		/*double x1;
		double x2;
		p->calcboundary(&x1, &x2);
		double avgpktsize = trgen->avgpktsize;
		double chiplength = (1.0/p->chiprate) * propSpeed;	
		double avgspan = avgpktsize*8*codeset->N*chiplength;	
		//Remove when the packet is three times beyond the 
		//zero point
		if (x2 < -1.0*3.0*avgspan)
		{
			dprintf("Intransit: traverse: Removing packet: %d\n", p->id);
	        	remove(p);
			if(current == NULL)
			{

				dprintf("Intransit: traverse: Found end.\n");
				p = current;
				current =  q->head;
				return(p);
			}
			
			p = current->next;
			current = current->next;
		}*/

	}
	return(p);
}
