/*
# Copyright (c) 2003-2004 by the University of Southern California.
# All rights reserved.
#
# Permission to use, copy, modify, and distribute this software and its
# documentation in source and binary forms for non-commercial purposes and
# without fee is hereby granted, provided that the above copyright notice
# appear in all copies and that both the copyright notice and this permission
# notice appear in supporting documentation, and that any documentation,
# advertising materials, and other materials related to such distribution and
# use acknowledge that the software was developed by the University of
# Southern California, Information Sciences Institute.  The name of the
# University may not be used to endorse or promote products derived from this
# software without specific prior written permission.
#
# THE UNIVERSITY OF SOUTHERN CALIFORNIA MAKES NO REPRESENTATIONS ABOUT THE
# SUITABILITY OF THIS SOFTWARE FOR ANY PURPOSE.  THIS SOFTWARE IS PROVIDED "AS
# IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT
# LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE.
#
# Other copyrights might apply to parts of this software and are so noted when
# applicable.
# ----------------------------------------------------------------------------
#
# Primary author: Purushotham Kamath
#                 Sept  2004
#This material is based upon work supported by the Defense Advanced Research 
#Projects Agency under contract no. N66001-02-1-8939 issued by the Space and 
#Naval Warfare Systems Center (SPAWAR).  Any opinions, findings, and conclusions 
#or recommendations expressed in this material are those of the author(s) and 
#do not necessarily reflect the views of the Defense Advanced Research Projects
#Agency, SPAWAR, or the U.S. Government.
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include <defs.h>
#include <packet.h>
#include <eventhandler.h>
#include <squeue.h>
#include <queue.h>
#include <scheduler.h>
#include <debug.h>
#include <intransit.h>
#include <state.h>
#include <codeset.h>
#include <defs.h>

extern Codeset * codeset;
extern int propSpeed;

/*
The State object has functions that calculate the state of the medium at a point in time, space
It uses the State object to determine the transmission times and location of packets.
State object is a singleton class
*/
extern Scheduler *sched;
extern InTransit *it;


//Section and PacketState classes are used to find the state within the boundaries of a packet.
//Used for implementing the receiver functionality


Section::Section(Packet *p)
{
	A = p->A;	
	N = p->N;

	q = new Queue(0);	

//Store a copy of the packet in the section
	Packet * pnew = new Packet(*p);	
	add(pnew);
	double thead, ttail;
	p->calcboundary(&thead, &ttail);
	head = thead; 
	tail = ttail;

 	next = NULL;
	prev = NULL;	
}

Section::~Section()
{
	Packet * p = q->head;
	while(p != NULL)
	{
		Packet *t = p->next;
		delete p;
		p = t; 
	}

}
Section::Section(Section & s)
{
	
	Packet *p = s.q->head;
	A = p->A;	
	N = p->N;



	q = new Queue(0);	
	while(p != NULL)
	{
		Packet *pnew = new Packet(*p); 
		q->enqueue(pnew);
		p = p->next;
	}
	
	next = s.next;
	prev = s.prev;


	x = s.x;
	y = s.y;
	z = s.z;
	head = s.head;
	tail = s.tail;

}

int Section::count()
{
	return(q->qNumber());

}

void Section::add(Packet *p)
{
	q->enqueue(p);	
}

Section * Section::insert(double x, double y, double z)
{

	Section *section = new Section(*this);	
	section->x = x;
	section->y = y;
	section->z = z;	

	section->head = x;
	section->tail = tail;
	tail = x;


	if(this->next != NULL)
		this->next->prev = section;

	section->next = this ->next ;
	this->next = section;
	section->prev = this;
	return(section);
}



void Section::dump()
{
	dprintf("state.cc: Section::dump(): Section head: %.16f\n", head); 		
	dprintf("state.cc: Section::dump(): Section tail: %.16f\n", tail); 		
	
	Packet *p=q->head;
	while(p != NULL)
	{
		dprintf("state.cc: Section::dump(): Packet id %d\n", p->id ); 		
		p = p->next;
	}


}


PacketState::PacketState(Packet * p)
{

	head  = new Section(p);


}

PacketState::~PacketState()
{
	Section *s = head;
	while(s != NULL)
	{
		Section *t = s->next;
		delete s;
		s = t; 
	}
}

void PacketState::dump()
{
	Section *section = head;
	while(section != NULL)
	{
		dprintf("Section:\n");
		section->dump();
		section = section->next;
	}	

}



void PacketState::ConstructStateSections(Packet *pthis)
{
	//Construct the state sections for the packet.
	dprintf("state.cc: PsacketState::ConstructStateSections(): Entered function\n");


	//The packet being received R consists of several sections
	//In each section the packet being received overlaps with a different set of other packets.
	//The steps are 
	//1. Identify the sections.
	//2. Calculate the state in each section
	//3. Calculate if the MUI in a section will cause loss of the packet

	//Identify the sections:
	//Determine all packets that overlap with this packet
	//These packets (P) must satisfy one of the following conditions:
	//Distances are distances from  a point beyond the head end where the packets are removed.

	//1. P completely overlaps R : Head(P) <= Head(R) and Tail(P) >= Head(R)   
	//2. R completely overlaps P : Head(P) > Head(R) and Tail(P) < Head(R)   
	//3. P partially overlaps R: Head(P) <= Head(R) and tail(P) < tail(R)
	//4. P partially overlaps R: Head(P) > Head(R) and tail(P) >= tail(R)


	//For each packet type split the receiving packet into sections


	double x1;
	double x2;
	pthis-> calcboundary(&x1, &x2);
		 

	//dprintf("trgen.cc: TrDump::handleevent(): Packet: %d\n",id);
	//dprintf("trgen.cc: TrDump::handleevent(): Section head: %.12f\n", rps->head->head);
	//dprintf("trgen.cc: TrDump::handleevent(): Section tail: %.12f\n", rps->head->tail);
	
	Packet *p;
	while((p = it->traverse()) != NULL )
	{

		if(pthis->id == p->id)
			continue;

		//dprintf("\n");
		//dprintf("trgen.cc: TrDump::handleevent(): Traversing packets on the line: %d\n", p->id);
		double packethead = 0.0;
		double packettail = 0.0;
		p->calcboundary(&packethead, &packettail);
	//	dprintf("trgen.cc: TrDump::handleevent(): Packet head: %.12f\n", packethead);
	//	dprintf("trgen.cc: TrDump::handleevent(): Packet tail: %.12f\n", packettail);

//State is measured from the point x to a point x-dx, where dx corresponds to the length of the state variable.
//Each element of the state array represents a chip on the line.
		//TBR
		//if(packettail < x2)
		//	continue;
//Case 1
		if ((packethead <= x1 ) && (packettail >= x2))
		{
		
			//P completely overlaps R - do not divide R
	//		dprintf("trgen.cc: TrDump::handleevent(): Complete overlap of P(%d) over R(%d).\n", p->id, id);
			Section *section = head;
			while(section!= NULL)
			{
				Packet *pnew = new Packet(*p);
				section->add(pnew);
				section = section->next;
			}
		}
		else if ((packethead > x1 ) && (packettail < x2))
		{
			//R completely overlaps P
	//		dprintf("trgen.cc: TrDump::handleevent(): Complete overlap of R(%d) over P(%d).\n", id, p->id);
			//Divide R into 3 sections
			Section *section = head;
			while(section!= NULL)
			{

				//5 cases to consider
				if ((section->head < packethead) && (section->tail <= packethead))
				{
					//Do nothing
					//dprintf("trgen.cc: TrDump::handleevent(): No overlap of section\n");
				}
				else if ((section->head  < packethead) && (section->tail <= packettail) && (section->tail) > packethead) 
				{
					//Split section	
					//dprintf("trgen.cc: TrDump::handleevent(): Partial overlap of section: type 1\n");
					Section * s = section->insert(packethead, 0, 0);
					Packet *pnew = new Packet(*p);
			 		s->add(pnew);
					section = s;							
				}		
				else if((section->head >= packethead ) && (section->head < packettail) && (section->tail > packettail)) 		
				{
					//dprintf("trgen.cc: TrDump::handleevent(): Partial overlap of section: type 2\n");
					//Split section	
					Section * s = section->insert(packettail, 0, 0);
					Packet *pnew = new Packet(*p);
			 		section->add(pnew);
					section = s;
				}
				else if((section->head >= packethead) && (section->tail <= packettail))
				{	
					//dprintf("trgen.cc: TrDump::handleevent(): Complete overlap of section by packet \n");
					Packet *pnew = new Packet(*p);
					section->add(pnew);	
				}
				else if((section->head < packethead ) && (section->tail > packettail))
				{
					//dprintf("trgen.cc: TrDump::handleevent(): Complete overlap of packet by section \n");
					
					Section * s = section->insert(packethead, 0, 0);
					Section * s2 = s->insert(packettail, 0, 0);
					Packet *pnew = new Packet(*p);
					s->add(pnew);	
					//dprintf("trgen.cc: TrDump::handleevent(): Dividing into 3 sections.\n", id, p->id);
					section = s2;
				}
				else if((section->head >= packettail) && (section->tail > packettail))
				{
					//Do nothing
					//dprintf("trgen.cc: TrDump::handleevent(): No overlap: Packet after section. \n");

				}
				section = section->next;
			}		

		}
		else if ((packethead < x1 ) && (packettail <= x2) && (packettail > x1))
		{
	//		dprintf("trgen.cc: TrDump::handleevent(): Partial overlap of  R(%d) and P(%d) - P starts before R.\n", id, p->id);
			Section *section = head;
			while(section!= NULL)
			{
				//Find the section to split
				if ((section->head >= packethead) && (section->tail <= packettail))
				{
					//Add to all the preceding packets
					//dprintf("trgen.cc: TrDump::handleevent(): Adding to section.\n");
					Packet *pnew = new Packet(*p);
					section->add(pnew);
				}
				else if ((section->head  < packettail) && (section->tail > packettail))
				{
					//Split section	
					//dprintf("trgen.cc: TrDump::handleevent(): Inserting section.\n");
					Section * s = section->insert(packettail, 0, 0);
					Packet *pnew = new Packet(*p);
			 		section->add(pnew);							
					section = s;	
				}
				else
				{	
					//Do nothing
				}				
				section = section->next;
			}			
		}

		else if ((packethead >= x1 ) && (packettail > x2) && (packethead < x2))
		{
	
	//		dprintf("trgen.cc: TrDump::handleevent(): Partial overlap of  R(%d) and P(%d) - R starts before P.\n", id, p->id);

			//Find the section to split
			Section *section = head;
			while(section!= NULL)
			{

				if ((section->head < packethead) && (section->tail <= packethead))
				{
					//Do nothing
					//dprintf("trgen.cc: TrDump::handleevent(): Nothing.\n");
				}
				else if ((section->head  < packethead) && (section->tail > packethead))
				{
					//Split section	
					//dprintf("trgen.cc: TrDump::handleevent(): Inserting section.\n");
					Section * s = section->insert(packethead, 0, 0);
					//dprintf("trgen.cc: TrDump::handleevent(): Adding packetyy.\n");
					Packet *pnew = new Packet(*p);
				 	s->add(pnew);					
					//dprintf("trgen.cc: TrDump::handleevent(): Added section.\n");
					section = s;		
				}
				else
				{	
					//dprintf("trgen.cc: TrDump::handleevent(): Adding to section.\n");
					//Add to all the following packets
					Packet *pnew = new Packet(*p);
					section->add(pnew);
				}				



				section = section->next;
				
			}


		}
		else
		{
			dprintf("trdump.cc: TrDump::Does not overlap()\n");\
			
		}

	}

	return;


}
