
/*
# Copyright (c) 2003-2004 by the University of Southern California.
# All rights reserved.
#
# Permission to use, copy, modify, and distribute this software and its
# documentation in source and binary forms for non-commercial purposes and
# without fee is hereby granted, provided that the above copyright notice
# appear in all copies and that both the copyright notice and this permission
# notice appear in supporting documentation, and that any documentation,
# advertising materials, and other materials related to such distribution and
# use acknowledge that the software was developed by the University of
# Southern California, Information Sciences Institute.  The name of the
# University may not be used to endorse or promote products derived from this
# software without specific prior written permission.
#
# THE UNIVERSITY OF SOUTHERN CALIFORNIA MAKES NO REPRESENTATIONS ABOUT THE
# SUITABILITY OF THIS SOFTWARE FOR ANY PURPOSE.  THIS SOFTWARE IS PROVIDED "AS
# IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT
# LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE.
#
# Other copyrights might apply to parts of this software and are so noted when
# applicable.
# ----------------------------------------------------------------------------
#
# Primary author: Purushotham Kamath
#                 Sept  2004
#This material is based upon work supported by the Defense Advanced Research 
#Projects Agency under contract no. N66001-02-1-8939 issued by the Space and 
#Naval Warfare Systems Center (SPAWAR).  Any opinions, findings, and conclusions 
#or recommendations expressed in this material are those of the author(s) and 
#do not necessarily reflect the views of the Defense Advanced Research Projects
#Agency, SPAWAR, or the U.S. Government.
*/


#ifndef STATE_H
#define STATE_H


/*
The State object has functions that calculate the state of the medium at a point in time, space
It uses the InTransit object to determine the transmission times and location of packets.


State is represented as an array of integers. The size of the array  is set before the simulation starts
in the main() function. The appropriate size for the state depends on several factors: The encoding used 
for the packets, how much state is needed by the application and how much state can be measured by the 
application.   
State object is a singleton class
*/


//The state of the packet
class Section 
{
public:
	int A;
	int N;

	Section *next;
	Section *prev;

//Coordinates of the starting point of the section
	double x;
	double y;
	double z;

	double head;
	double tail;

	Queue *q;
	Section(Packet *);
	~Section();
	Section(Section &);
	
	int count();
	void add(Packet * p);
	Section * insert(double, double, double);
	void dump();
};

//The state of the packet
class PacketState 
{
public:
	Section * head;
	
	PacketState(Packet *);
	virtual ~PacketState();
	void ConstructStateSections(Packet *);
	void dump();
		
};

#endif

